#include "global.h"

#include "Port.h"
#include <Preferences.h> //lib pour écrire dans l'eeprom
//#include <arpschuino32_core.h>
#include <arpschuino.h>
#include "custom_port.h"

//#include <AccelStepper.h>
#include "ArpAccelStepper.h"

String autocalibration_status = "autocalibration request";

#ifdef WILULU32
    Port portW('W');
#endif
Port portA('A'); // creation des objets port
#ifdef ARPSCHUINO32
    Port portB('B');
#endif

Preferences port_preferences;

Port::Port(char name) : m_port_name(name)
{

}

///////////////////////////////////////////////////////
////////////////OTHER METHODS/////////////////////////
//////////////////////////////////////////////////////
void Port::find_mode()
{
    port_preferences.begin("port_datas", false);
    uint8_t defaultVal = 0;
    m_mode = load("mode", defaultVal); 
}

void Port::debug()
{
    Serial.print("debug port ");
    Serial.println(m_port_name);
}
///////////////////////////////////////////////////////
////////////////SET METHODS////////////////////////////
//////////////////////////////////////////////////////

void Port::setNbDmxChannels(uint16_t nbChannels) // should auto adjust, except for pixel Led
{
    m_nbDmxChannels = nbChannels;
}
///////////////////////////////////////////////////////
////////////////GET METHODS////////////////////////////
//////////////////////////////////////////////////////

uint16_t Port::getNbDmxChannels() const
{
    return m_nbDmxChannels;
}

uint16_t Port::getMode() const
{
    return m_mode;
}

String Port::writeMode() const
{
    String str;
    switch (getMode())
    {
    case 0:
        str = "PWM";
        break;
    case 1:
        str = "on/off";
        break;
    case 2:
        str = "servo";
        break;
    case 3:
        str = "stepper";
        break;
    case 4:
        str = "CCmotor";
        break;
    case 20:
        str = "inputs";
        break; 
    case 254:
        str = "custom";
        break;               
    default:
        str = "inactive";
    }
    return str;
}
/////////////////////////////////////////////////////////
////////////////RECORD METHODS//////////////////////////
///////////////////////////////////////////////////////
void Port::record(String name, uint8_t val) const
{
    String str(m_port_name);
    str += name;
    const char *key = str.c_str();
    port_preferences.putUChar(key, val);
}

void Port::record(String name, uint16_t val) const
{
    String str(m_port_name);
    str += name;
    const char *key = str.c_str();
    port_preferences.putUShort(key, val);
}

void Port::record(String name, uint32_t val) const
{
    String str(m_port_name);
    str += name;
    const char *key = str.c_str();
    port_preferences.putULong(key, val);
}

void Port::record(String name, bool val) const
{
    String str(m_port_name);
    str += name;
    const char *key = str.c_str();
    port_preferences.putBool(key, val);
}

void Port::record(String name, float val) const
{
    String str(m_port_name);
    str += name;
    const char *key = str.c_str();
    port_preferences.putFloat(key, val);
}
/////////////////////////////////////////////////////////////////////////////
//////////////////////READ METHODS (read preferences)/////////////////////
////////////////////////////////////////////////////////////////////////////

uint8_t Port::load(String name, uint8_t val) const
{
    String str(m_port_name);
    str += name;
    const char *key = str.c_str();
    return port_preferences.getUChar(key, val);
}

uint16_t Port::load(String name, uint16_t val) const
{
    String str(m_port_name);
    str += name;
    const char *key = str.c_str();
    return port_preferences.getUShort(key, val);
}

uint32_t Port::load(String name, uint32_t val) const
{
    String str(m_port_name);
    str += name;
    const char *key = str.c_str();
    return port_preferences.getULong(key, val);
}

bool Port::load(String name, bool val) const
{
    String str(m_port_name);
    str += name;
    const char *key = str.c_str();
    return port_preferences.getBool(key, val);
}

float Port::load(String name, float val) const
{
    String str(m_port_name);
    str += name;
    const char *key = str.c_str();
    return port_preferences.getFloat(key, val);
}

///////////////////////////////////////////////////////
//////////////////SETUP METHODS////////////////////////
///////////////////////////////////////////////////////
void Port::init()
{
    #ifdef WILULU32
        m_nb_Pins = (m_port_name == 'W') ? NB_WILULU_OUT : 8;
    #endif
    #ifdef ARPSCHUINO32
        m_nb_Pins = 8;
    #endif    
    m_startPin = (m_port_name == 'A') ? 0 : 8;/////////////////////////////////
    Serial.print("Port ");
    Serial.print(m_port_name);
    Serial.print(": ");Serial.print(writeMode());
    
    //
    
    Serial.println(" ");
    switch (m_mode)
    {
        case 0:
        { // pwm
            PWM_setup();
            break;
        }
        case 1:
        {
            on_off_setup();
            break;
        }
        case 2:
        {
            servo_setup();
            break;
        }
        case 3:
        {
            stepper_setup();
            break;
        }
        case 4:
        {
            CCmotor_setup();
            break;
        }
        case 20:
        {
            inputs_setup();
            break;
        } 
        case 254:
        {
            custom_setup();
            break;
        }
        default:
        {
            //Serial.println(": inactive...");
        } 
        
    }
    //Serial.print(" m_nbDmxChannels : ");Serial.print(m_nbDmxChannels);
    Serial.println(); 
}

void Port::PWM_setup()
{
    int32_t freq = 40000000;
    uint8_t defaultVal = 14;
    m_pwmRes = load("PWM_resolution", defaultVal);////////////////////////////////////////////////////////
    //Serial.print("la resolutions est :");Serial.println(m_pwmRes);
    for (uint8_t i = 1; i < m_pwmRes; i++)
    {
        freq = freq / 2;
    }
    //uint8_t startPin = (m_port_name == 'A') ? 0 : 8;
    m_dmxRes = load("dmxRes", false);
    if (m_dmxRes)
    {
        m_nbDmxChannels = m_nb_Pins*2;
    }
    else
    {
        m_nbDmxChannels = m_nb_Pins;
    }
    for (uint8_t i = m_startPin; i < m_startPin + m_nb_Pins; i++)
    {
        // pinMode(Arp[i], OUTPUT);
        ledcSetup(i, freq, m_pwmRes); // configure LED PWM functionalitites
        ledcAttachPin(Arp[i], i);     // attach the channel to the GPIO to be controlled
    }
    //Serial.print(": PWM ");
    Serial.print(m_pwmRes);
    Serial.print(" bits; freq ");
    Serial.print(freq);
    Serial.print(" hz");
    if (m_dmxRes)
        Serial.print(", 16 bits DMX input");
    Serial.println();
}

void Port::on_off_setup()
{
    m_inverted = load("inverted", false); // defaultVal);
    uint8_t defaultVal = 128;
    m_trigger = load("trigger", defaultVal);
    //uint8_t startPin = (m_port_name == 'A') ? 0 : 8;
    m_nbDmxChannels = m_nb_Pins;
    for (int i = m_startPin; i < m_startPin + m_nb_Pins; i++)
    {
        pinMode(Arp[i], OUTPUT);
        if (m_inverted)
            digitalWrite(Arp[i], HIGH);
    }
    Serial.print(" trigger at ");
    Serial.print(m_trigger);
    Serial.print(" (");
    Serial.print((m_trigger * 100) / 255);
    Serial.print("%)");
    if (m_inverted)
        Serial.print(", inverted");
    Serial.println();
}

void Port::servo_setup()
{
    m_inverted = load("inverted", false);
    uint8_t defaultAngle = 160;
    m_servoAngle = load("angle", defaultAngle);
    m_servoDetach = load("detach", false);
    m_nbDmxChannels = 8;
    // uint8_t startPin= 0 ;
    //uint8_t startPin = (m_port_name == 'A') ? 0 : 8;
    m_servo = new Servo[8];
    if (m_servoDetach == false)
    {
        for (int i = 0; i < 8; i++)
        {
            uint8_t output = i + m_startPin;
            m_servo[i].attach(Arp[output], output); //(pin,channel)
        }
    }
    Serial.print(": Servo");
    Serial.print(", ");
    Serial.print(m_servoAngle);
    Serial.print("°");
    if (m_inverted)
        Serial.print(", inverted");
    if (m_servoDetach)
        Serial.print(", mode detach");
    Serial.println();
}

void Port::stepper_setup()
{
    // default values
    uint16_t defaultSPR = 800; // step per revolution
    uint16_t defaultMin = 4;   // speed mini in RPM
    uint16_t defaultMax = 200; // speed maxi in RPM
    float defaultRevol = 12;   // nb of revolution
    float defaultAccel = 500;//   Acceleration rate ////////////////////////
    bool defaultAction = 0;    // perform by default
    uint8_t default_input =255;

    //uint8_t m_startPin = (m_port_name == 'A') ? 0 : 8;
    m_nbDmxChannels = 8;

    m_stepper = new ArpAccelStepper[4];
    bool homing[4]={false};

    for (int i = 0; i < 4; i++)
    {
        String num(i);
        String str = num + "stpActive";
        const char *word = str.c_str();
        m_stp_active[i] = load(word, true);
        Serial.print("stepper ");
        Serial.print(i);
        if (m_stp_active[i])
        {
            str = num + "stpPerRevol";
            word = str.c_str();
            uint16_t stepPerRevolution = load(word, defaultSPR);
            m_stepper[i].set_interface(ArpAccelStepper::DRIVER);//permettre d'autre types de drivers ?
            m_stepper[i].attach(Arp[i * 2 + 1 + m_startPin], Arp[i * 2 + m_startPin], stepPerRevolution);
            m_stepper[i].setMinPulseWidth(7);

            Serial.print(" steps per revolution: ");
            Serial.print(stepPerRevolution);
            Serial.print("; ");
            
            str = num + "_rpm_min";
            word = str.c_str();
            uint16_t rpmMin = load(word, defaultMin);
            Serial.print(" rpmmin: ");
            Serial.print(rpmMin);
            Serial.print("; ");

            str = num + "_rpm_max";
            word = str.c_str();
            uint16_t rpmMax = load(word, defaultMax);
            Serial.print(" max: ");
            Serial.print(rpmMax);
            Serial.print("; ");

            str = num + "_revol";
            word = str.c_str();
            float revol = load(word, defaultRevol);
            Serial.print(" revolutions: ");
            Serial.print(revol);
            Serial.print("; ");

            m_endPosition[i] = stepPerRevolution * revol;

            str = num + "_accel";//////////////////////////////////////////////////
            word = str.c_str();
            float acceleration = load(word, defaultAccel);
            Serial.print(" acceleration: ");
            Serial.print(acceleration);
            Serial.print("; ");            

            str = num + "stpAction";
            word = str.c_str();
            m_action[i] = load(word, defaultAction);//variable action dans port.h
            Serial.print(" action: ");
            uint16_t dmxRes = _8bits;

            if (m_action[i] == 0)
            {
                Serial.print("perform");
                Serial.print("; ");
                Serial.print(" DMX resolution: ");
                str = num + "stp_dmxRes";
                word = str.c_str();
                m_stp_dmxRes[i] = load(word, false);
                if (m_stp_dmxRes[i] == true)
                {
                    dmxRes = _16bits;
                    Serial.print("16bits");
                    m_nbDmxChannels++;
                }
                else
                {
                    Serial.print("8bits");
                }
                Serial.print("; ");
            }
            else
            {
                Serial.print("continuous; ");
            }
            m_stepper[i].init(rpmMin, rpmMax, revol, dmxRes,m_action[i],acceleration);
            //init(int16_t RPM_min=6,int16_t RPM_max=300,float revolutions=10,uint16_t resolution=255,bool action=0,float acceleration=10000);
            str = num + "stpInverted";
            word = str.c_str();
            bool inverted = load(word, false);
            if (inverted)
            {
                m_stepper[i].invert_rotation(true);
                Serial.println(" inverted; ");
            }
            str = num + "stpZeroSW";
            word = str.c_str();
            m_stpZeroSW[i] = load(word, false);
            //Serial.print(m_CCZeroSW[i]? ", Zero switch":"");

            uint8_t homSpeed;

            if(m_stpZeroSW[i])
            {
                Serial.print(", Zero switch");
                str = num + "stpZero_i";
                word = str.c_str();
                m_stpzero_input[i] = load(word, default_input);
                if(m_stpzero_input[i]<255)
                {
                    if (m_stpzero_input[i]<16)
                        pinMode(Arp[m_stpzero_input[i]],INPUT_PULLUP);
                    else if(m_stpzero_input[i]<20)
                        pinMode(Arp[m_stpzero_input[i]],INPUT);
                    Serial.print(": Arp");
                    Serial.print(m_stpzero_input[i]);
                    Serial.print("; ");
                }
                else Serial.print(": no pin assigned"); 

                str = num + "stpHomSpeed";
                word = str.c_str();
                homSpeed = rpmMax/load(word, (uint8_t)4);


                str = num + "stpHomingSW";
                word = str.c_str();
                homing[i] = load(word, false);
                if(homing[i])
                {
                    Serial.print(" homing at startup; ");
                    m_stepper[i].setMaxSpeed(m_stepper[i].RPMtoSPS (homSpeed));//speed for homing
                }
            }

            str = num + "stpEndSW";
            word = str.c_str();
            m_stpEndSW[i] = load(word, false);
            //Serial.print(m_CCEndSW[i]? ", End switch":"");           
            if(m_stpEndSW[i])
            {
                Serial.print(", End switch");
                str = num + "stpEnd_i";
                word = str.c_str();
                m_stpend_input[i] = load(word, default_input);
                if(m_stpend_input[i]<255)
                {
                    if (m_stpzero_input[i]<16)
                        pinMode(Arp[m_stpend_input[i]],INPUT_PULLUP);
                    else if(m_stpzero_input[i]<20)
                        pinMode(Arp[m_stpend_input[i]],INPUT);
                    //attachInterrupt(m_stpzero_input[i], fonction_ISR, FALLING);///////////////////////////////////////

                    Serial.print(": Arp");
                    Serial.print(m_stpend_input[i]);
                }
                else Serial.print(": no pin assigned"); 
            }                   
            Serial.println();
            ///////////////// autocalibration /////////////////////////////
            str = num + "autocalib";//stepperNum + "autocalibration"
            word = str.c_str();
            bool default_val=false;
            bool do_autocal = load(word, default_val);
            //Serial.println(do_autocal);
         
            if(do_autocal)
            {
                Serial.print("autocalibrating ");
                record(word, false);

                m_stepper[i].setMaxSpeed(m_stepper[i].RPMtoSPS (homSpeed));//speed for homing
                m_stepper[i].move(0x6FFFFFFF);//go to the end
                autocalibration_status = "go to the end switch";
                while(digitalRead(Arp[m_stpend_input[i]]))//end switch not pressed
                {   
                    // Print dots while waiting for homing to finish
                    static uint16_t dot=0;
                    if(dot==0)  
                        Serial.print('.');
                    dot++;
                    ////////////////////////////////////////////////
                    m_stepper[i].run(); 
                }
                Serial.print("stepper");Serial.print(i);Serial.print(" at end !");
                m_stepper[i].setCurrentPosition(0);
                m_stepper[i].move(-0x6FFFFFFF); //go back
                autocalibration_status = "back to the zero switch";
                while(digitalRead(Arp[m_stpzero_input[i]]))//zero switch not pressed
                {   
                    // Print dots while waiting for homing to finish
                    static uint16_t dot=0;
                    if(dot==0)  
                        Serial.print('.');
                    dot++;
                    ////////////////////////////////////////////////
                    m_stepper[i].run(); 
                }
                Serial.print("stepper");Serial.print(i);Serial.println(" at zero !");
                m_endPosition[i] = abs(m_stepper[i].currentPosition());
                m_stepper[i].setCurrentPosition(0);
                autocalibration_status = "autocalibration complete";
                Serial.print("autocalibrating complete, ");
                revol = ((float)m_endPosition[i]/stepPerRevolution);

                str = num + "_revol";
                word = str.c_str();
                record(word, revol);
                Serial.print("new revolutions value: ");
                Serial.println(revol);
                //Serial.print("; ");
                m_stepper[i].init(rpmMin, rpmMax, revol, dmxRes,m_action[i],acceleration);
            }

        }
        else
        {
            Serial.println(" inactive; ");
            //Serial.println();
        }        
    } 
    Serial.println();
    /////////////////////////////////////// homing ////////////////////////
    if(homing[0] || homing[1] || homing[2] || homing[3])
    {
        Serial.print("homing port ");Serial.print(m_port_name);
        bool atHome[4]={false};      
        while(not atHome[0] || not atHome[1] || not atHome[2] || not atHome[3])
        {   
            // Print dots while waiting for homing to finish
            static uint16_t dot=0;
            if(dot==0)  
                Serial.print('.');
            dot++;

            for(int i=0;i<4;i++)
            {
                if(homing[i] && not atHome[i])
                {
                    m_stepper[i].move(-0x6FFFFFFF); 
                    if(digitalRead(Arp[m_stpzero_input[i]]))//switch not pressed
                    {
                        m_stepper[i].run();
                    }    
                    else //switch pressed
                    {
                        Serial.print("stepper");Serial.print(i);Serial.print(" at home !");
                        m_stepper[i].setCurrentPosition(0);
                        atHome[i]=true;
                    }                    
                    
                }
                else atHome[i]=true;
            }
        }
    }
    Serial.println();




}

void Port::CCmotor_setup()
{
    uint8_t default_CC_driver_type = 0;
    uint8_t default_CC_threshold = 1;
    uint8_t default_input =255;
    //uint8_t startPin = (m_port_name == 'A') ? 0 : 8;
    m_nbDmxChannels = m_nb_Pins;

    for (int i = 0; i < 4; i++)
    {
        String num(i);
        String str = num + "CCActive";
        const char *word = str.c_str();
        m_CC_active[i] = load(word, true);
        Serial.print("DC motor ");
        Serial.print(i);
        if (m_CC_active[i])
        {
            //m_CC_driver_type[i] = 1;//new : 0=Hbridge classic, 1=speed/dir
            str = num + "CCdvrType";//limited to a maximum of 15 characters (so 10 !)
            word = str.c_str();
            m_CC_driver_type[i] = load(word, default_CC_driver_type);

            Serial.print((m_CC_driver_type[i] == 0) ? " H-bridge classic" : " speed/dir");            

            str = num + "thresold";
            word = str.c_str();
            m_cc_threshold[i] = load(word, default_CC_threshold);
            Serial.print(", thresold: ");
            Serial.print(m_cc_threshold[i]);

            uint8_t percent = m_cc_threshold[i] * 100 / 255;
            Serial.print(" (");
            Serial.print(percent);
            Serial.print("%)");

            str = num + "CCInverted";
            word = str.c_str();
            m_CC_inverted[i] = load(word, false);
            Serial.print(m_CC_inverted[i]? ", inverted":"");

            str = num + "CCZeroSW";
            word = str.c_str();
            m_CCZeroSW[i] = load(word, false);
            //Serial.print(m_CCZeroSW[i]? ", Zero switch":"");
            if(m_CCZeroSW[i])
            {
                Serial.print(", Zero switch");
                str = num + "CCZero_i";
                word = str.c_str();
                m_CCzero_input[i] = load(word, default_input);
                if(m_CCzero_input[i]<255)
                {
                    pinMode(Arp[m_CCzero_input[i]],INPUT_PULLUP);
                    Serial.print(": Arp");
                    Serial.print(m_CCzero_input[i]);
                }
                else Serial.print(": no pin assigned");           
            }

            str = num + "CCEndSW";
            word = str.c_str();
            m_CCEndSW[i] = load(word, false);
            //Serial.print(m_CCEndSW[i]? ", End switch":"");           
            if(m_CCEndSW[i])
            {
                Serial.print(", End switch");
                str = num + "CCEnd_i";
                word = str.c_str();
                m_CCend_input[i] = load(word, default_input);
                if(m_CCend_input[i]<255)
                {
                    pinMode(Arp[m_CCend_input[i]],INPUT_PULLUP);
                    Serial.print(": Arp");
                    Serial.print(m_CCend_input[i]);
                }
                else Serial.print(": no pin assigned"); 
            }       
        }
        else
        {
            Serial.print(" inactive");
        }
        Serial.println(); 
    }



    for (int i = m_startPin; i < m_startPin + 8; i++)
    {
        ledcSetup(i, 25000, 8);   // channel, freq, Resolution,25000Hz should be unaudible
        ledcAttachPin(Arp[i], i); // attach the channel to the GPIO to be controlled
    }
}

void Port::inputs_setup()
{
    //uint8_t startPin = (m_port_name == 'A') ? 0 : 8;
    m_nbDmxChannels = 0;
    for (int i = m_startPin; i < m_startPin + 8; i++)
    {
        pinMode(Arp[i], INPUT_PULLUP);
    }
    Serial.println();
}

void Port::custom_setup()
{
    #ifdef WILULU32
        if (m_port_name == 'W')
        {
            setupW();
        }
    #endif    
    if (m_port_name == 'A')
    {
        setupA();
    }
    #ifdef ARPSCHUINO32
        else if (m_port_name == 'B')
        {
            setupB();
        }
    #endif
}

///////////////////////////////////////////////////////
//////////////////ACTION METHODS////////////////////////
///////////////////////////////////////////////////////
void Port::action(const std::vector<uint8_t> &DMXslice)
{
    //uint8_t startPin = (m_port_name == 'A') ? 0 : 8;
                    //Serial.print("ch1 ");Serial.print(DMXslice[0]);Serial.print(" ch2");Serial.println(DMXslice[1]);
    //Serial.println(DMXslice[0]);
    switch (m_mode)
    {
        case 0:
        {
            PWM_action(DMXslice);
            break;
        }
        case 1:
        {
            on_off_action(DMXslice);
            break;
        }
        case 2:
        {
            servo_action(DMXslice);
            break;
        }
        case 3:
        {
            stepper_action(DMXslice);
            break;
        }
        case 4:
        {
            CCmotor_action(DMXslice);
            break;
        }
        case 254:
        {
            custom_action(DMXslice);
        }
    }
}

void Port::PWM_action(const std::vector<uint8_t> &DMXslice)
{
    if (m_dmxRes)
    { // if 16 bits DMX
        if (m_pwmCurve == 0)
        { // if led curve
            for (int i = 0; i < m_nb_Pins; ++i)
            {
                m_target_level[i] = DMXslice[i * 2] << 8 | DMXslice[i * 2 + 1];             // level= DMXslice[i*2] << 8 | DMXslice[i*2+1];
                uint32_t level32 = ((m_target_level[i] + 1) * (m_target_level[i] + 1)) - 1; // uint32_t level32 =((level+1)*(level+1))-1;
                m_target_level[i] = level32 >> (32 - m_pwmRes);                             // level = level32 >>(32-m_pwmRes);
            }
        }
        else
        { // if straight curve
            for (int i = 0; i < m_nb_Pins; ++i)
            {
                m_target_level[i] = DMXslice[i * 2] << 8 | DMXslice[i * 2 + 1]; // level= DMXslice[i*2] << 8 | DMXslice[i*2+1];
            }
        }
    }
    else
    { // if 8 bits DMX
        if (m_pwmCurve == 0)
        { // if led curve
            for (int i = 0; i < m_nb_Pins; ++i)
            {
                m_target_level[i] = (((DMXslice[i] + 1) * (DMXslice[i] + 1) - 1) >> (16 - m_pwmRes)); // level = (((DMXslice[i]+1)*(DMXslice[i]+1)-1)>>(16-m_pwmRes));
            }
        }
        else
        { // if straight curve
            for (int i = 0; i < m_nb_Pins; ++i)
            {
                m_target_level[i] = DMXslice[i] << (m_pwmRes - 8); // level= DMXslice[i]<<(m_pwmRes-8);
            }
        }
    }
    //Serial.println(m_target_level[0]);
    //Serial.println(sizeof(DMXslice));//DMXslice[0]);
    //Serial.println(DMXslice[0]);
}

void Port::on_off_action(const std::vector<uint8_t> &DMXslice) const
{
    //Serial.println(DMXslice[0]);//////////////////////////////////////////////////////////
    for (int i = 0; i < m_nbDmxChannels; ++i)
    {
        if (DMXslice[i] > m_trigger - 1)
        {
            digitalWrite(Arp[m_startPin + i], 1 - m_inverted);
        }
        else
        {
            digitalWrite(Arp[m_startPin + i], m_inverted);
        }
    }
    Serial.print("on off");
}

void Port::servo_action(const std::vector<uint8_t> &DMXslice) const
{
    Serial.println(DMXslice[0]);
    uint8_t min = 0, max = m_servoAngle;
    if (m_inverted)
    {
        min = m_servoAngle, max = 0;
    }
    if (m_servoDetach)
    {
        static uint8_t oldVal[8] = {0};
        for (int i = 0; i < m_nbDmxChannels; ++i)
        {
            if (DMXslice[i] != oldVal[i])
            {
                uint8_t output = i + m_startPin;
                m_servo[i].attach(Arp[output], output); //(pin,channel)
                m_servo[i].write(map(DMXslice[i], 0, 255, min, max));
            }
            else if (m_servo[i].attached())
            {
                m_servo[i].detach();
            }
            oldVal[i + m_startPin] = DMXslice[i];
        }
    }
    else
    {
        for (int i = 0; i < m_nbDmxChannels; ++i)
        {
            m_servo[i].write(map(DMXslice[i], 0, 255, min, max));
        }
    }
}

void Port::stepper_action(const std::vector<uint8_t> &DMXslice) const
{
    uint16_t this_stepper_addr = 0;
    for (int i = 0; i < 4; ++i)
    {
        if (m_stp_active[i])
        {            
            if (m_stp_dmxRes[i]) // 16 Bits
            {
                uint16_t position = ((DMXslice[this_stepper_addr] << 8) | DMXslice[this_stepper_addr + 1]);
                m_stepper[i].refresh(position, DMXslice[this_stepper_addr + 2],m_action[i]);
                this_stepper_addr = this_stepper_addr + 3;
            }
            else // 8 Bits
            {
                m_stepper[i].refresh(DMXslice[this_stepper_addr], DMXslice[this_stepper_addr + 1],m_action[i]);
                this_stepper_addr = this_stepper_addr + 2;
            }
        }
    }
}

void Port::CCmotor_action(const std::vector<uint8_t> &DMXslice) const
{
    uint16_t this_CC_addr = 0;
    for (int i = 0; i < 4; i++)
    {
        if (m_CC_active[i])
        {
            // uint8_t state = DMXslice[this_CC_addr];
            // bool direction = (state > 127) ? 0 : 1;
            // direction = direction - m_CC_inverted[i];
            uint8_t speed = DMXslice[this_CC_addr + 1];

            //Serial.print(i);Serial.print(" : ");Serial.println(m_CC_driver_type[i]);

            if (speed > 0)
            {
                speed = map(speed, 1, 255, m_cc_threshold[i], 255); // map(value, fromLow, fromHigh, toLow, toHigh)
                //to be optimised
                //uint8_t state = DMXslice[this_CC_addr];
                bool direction = (DMXslice[this_CC_addr] > 127) ? 0 : 1;
                direction = direction - m_CC_inverted[i];



                // switch (m_CC_driver_type[i])
                // {
                //     case 0://0=Hbridge classic
                //     { 
                        if (direction) // avant
                        {
                            if(m_CCEndSW[i])
                            {
                                if(!digitalRead(Arp[m_CCend_input[i]]))//end switch pressed
                                    speed=0;
                            }
                            if(m_CC_driver_type[i]==0)//H-bridge classic
                            {
                                ledcWrite(m_startPin + i * 2, speed);
                                ledcWrite(m_startPin + i * 2 + 1, 0);                                
                            }                            
                            // else if(m_CC_driver_type[i]==1)//speed/dir driver
                            // {
                            //     ledcWrite(m_startPin + i * 2, 255);
                            //     ledcWrite(m_startPin + i * 2 + 1, speed);
                            // }
                            else if(m_CC_driver_type[i]==1)//speed/dir driver
                            {
                                ledcWrite(m_startPin + i * 2, 255);//DIR
                                ledcWrite(m_startPin + i * 2 + 1, speed);
                            }                            
                        }
                        else // arrière
                        {
                            if(m_CCZeroSW[i])
                            {
                                if(!digitalRead(Arp[m_CCzero_input[i]]))//zero switch pressed
                                    speed=0;
                            } 
                            ledcWrite(m_startPin + i * 2, 0);
                            ledcWrite(m_startPin + i * 2 + 1, speed);                            
                            // if(m_CC_driver_type[i]==0)
                            // {                                   
                            //     ledcWrite(m_startPin + i * 2, 0);
                            //     ledcWrite(m_startPin + i * 2 + 1, speed);
                            // }
                            // else if(m_CC_driver_type[i]==1)
                            // {
                            //     ledcWrite(m_startPin + i * 2, speed);
                            //     ledcWrite(m_startPin + i * 2 + 1, 0);
                            // }                            
                        }
                        // break;
                //     }
                //     case 1:// 1=dir/speed
                //     {
                //         if (direction) // avant
                //         { 
                //             if(m_CCEndSW[i])
                //             {
                //                 if(!digitalRead(Arp[m_CCend_input[i]]))//end switch pressed
                //                     speed=0;
                //             }
                //             ledcWrite(m_startPin + i * 2, 255);
                //             ledcWrite(m_startPin + i * 2 + 1, speed);                                                         
                //         }
                //         else // arrière
                //         {
                //             if(m_CCZeroSW[i])
                //             {
                //                 if(!digitalRead(Arp[m_CCzero_input[i]]))//zero switch pressed
                //                     speed=0;
                //             }                         
                //         }
                //             ledcWrite(m_startPin + i * 2, 0);
                //             ledcWrite(m_startPin + i * 2 + 1, speed); 
                //         break;                       
                //     }
                //     default:
                //     {
                //         Serial.print("error : invalid CC_driver_type");//Serial.println(m_CC_driver_type[i]);
                //     } 
                // }
            }

            else //if (m_CC_driver_type[i]==0)// Stop (freinage)
            {
                ledcWrite(m_startPin + i * 2, 0);
                ledcWrite(m_startPin + i * 2 + 1, 0);
            }
            
            this_CC_addr = this_CC_addr + 2;
        }
    }
    //Serial.println(m_nbDmxChannels);//(sizeof(DMXslice));
    //Serial.println(DMXslice[0]);
}

void Port::custom_action(const std::vector<uint8_t> &DMXslice) const
{
    #ifdef WILULU32
        if (m_port_name == 'W')
        {
            action_W(DMXslice);
        }
    #endif
    if (m_port_name == 'A')
    {
        action_A(DMXslice);
    }
    #ifdef ARPSCHUINO32
        else if (m_port_name == 'B')
        {
            action_B(DMXslice);
        }
    #endif
}

void Port::runAtEachLoop()
{
    switch (m_mode)
    {
        case 0:
        { // PWM
            for (uint8_t i = 0; i < m_nb_Pins; i++)
            {
                int32_t level;
                if (m_actual_level[i] < m_target_level[i])
                {
                    level = m_actual_level[i] + 1 + (m_actual_level[i] / 256); // m_actual_level[i]++;
                    if (level > m_target_level[i])
                        m_actual_level[i] = m_target_level[i];
                    else
                        m_actual_level[i] = level;
                    ledcWrite(m_startPin + i, m_actual_level[i]);
                }
                else if (m_actual_level[i] > m_target_level[i])
                {
                    level = m_actual_level[i] - 1 - (m_actual_level[i] / 256); // m_actual_level[i]--;
                    if (level < m_target_level[i])
                        m_actual_level[i] = m_target_level[i];
                    else
                        m_actual_level[i] = level;
                    ledcWrite(m_startPin + i, m_actual_level[i]);
                }
            }
            break;
        }
        case 3:
        { // stepper
            for (uint8_t i = 0; i < 4; i++)
            {
                if (m_stp_active[i])
                {
                    if(m_stpZeroSW[i] && m_stepper[i].distanceToGo()<0)
                    {
                        if(digitalRead(Arp[m_stpzero_input[i]]))
                            m_stepper[i].run();
                        else m_stepper[i].setCurrentPosition(0);
                    }     
                    else if(m_stpEndSW[i] && m_stepper[i].distanceToGo()>0)
                    {
                        if(digitalRead(Arp[m_stpend_input[i]]))
                            m_stepper[i].run();
                        else m_stepper[i].setCurrentPosition(m_endPosition[i]);
                    }
                    else m_stepper[i].run();                    
                }
            }
            break;
        }
        case 4:
        { // CC motor
            for (uint8_t i = 0; i < 4; i++)
            {
                if (m_CC_active[i])
                {
                    switch (m_CC_driver_type[i])
                    {
                        case 0://0=Hbridge classic
                        {
                            if(m_CCZeroSW[i])// && ledcRead(m_startPin+i))//ledcRead(uint8_t chan)
                            {
                                if(!digitalRead(Arp[m_CCzero_input[i]]))    //zero switch pressed
                                {
                                    ledcWrite(m_startPin + i * 2 + 1, 0);   //stop motor if backward
                                    //Serial.print ("!");
                                } 
                            }       
                            if(m_CCEndSW[i])// && !ledcRead(m_startPin+i))
                            {
                                if(!digitalRead(Arp[m_CCend_input[i]]))     //zero switch pressed
                                {
                                    ledcWrite(m_startPin + i * 2, 0);       //stop motor if forward
                                    //Serial.print ("?");
                                }   
                            }
                            break;
                        }
                        case 1:
                        {
                            if((m_CCZeroSW[i] && ledcRead(m_startPin+i))||(m_CCEndSW[i] && !ledcRead(m_startPin+i)))//ledcRead(uint8_t chan)
                            {
                                ledcWrite(m_startPin + i * 2 + 1, 0);   //stop motor
                            } 
                            break;
                        }
                    } 
                }                 
            }
            break;
        }    
        case 254:
        {
            #ifdef WILULU32
                if (m_port_name == 'W')
                {
                    at_each_loop_W();
                }
            #endif        
            if (m_port_name == 'A')
            {
                at_each_loop_A();
            }
            #ifdef ARPSCHUINO32
                else if (m_port_name == 'B')
                {
                    at_each_loop_B();
                }
            #endif
            break;
        }
        default : 
        {
            //do nothing...
            //Serial.println("dflt");//debug
        }
    }
}


///////////////////////////////////////// interruptions ////////////////////////////////
// void IRAM_ATTR Port:: fonction_ISR() {
//     ets_printf("Boutton pressé\n");
//     // Code de la fonction
// }

// void IRAM_ATTR Port:: fonction_ISR2() {
//     ets_printf("Boutton relaché\n");
//     // Code de la fonction
// }

void Port::stpHome(uint8_t stpNum, uint8_t speed)
{

}